/*
	KUMAamp project()
	Copyright (C) Mamiya 2000.
	License:LGPL
	---------------------------------------------------------------------------
	NSF song Controler
	con_nsf.c
*/

#include "in_mdx.h"

/* ANSI/Windows standard headers */
#include <windows.h>
#include <shlobj.h>

/* Libraries headers */
#ifdef __cplusplus
extern "C" {
#endif
#include "../winamp/in2.h"
#include "../winamp/wafe.h"
#ifdef __cplusplus
}
#endif

/* Project headers */
#include "con_nsf.h"
#include "cso.h"

#ifdef __cplusplus
extern "C" In_Module __declspec(dllexport) * winampGetInModule2(void);
#else
extern In_Module __declspec(dllexport) * winampGetInModule2(void);
#endif


static struct
{
	BOOL enable;
	int songmax;
	int songno;
	HWND hwndWA;
	WNDPROC wpOrigWA;
	HWND hwndEQ;
	WNDPROC wpOrigEQ;
	HWND hwndPE;
	WNDPROC wpOrigPE;
} scwa = { FALSE, 1, 1, NULL };
static CRITICAL_SECTION_OBJECT(scwa_lock);
typedef enum {
	SUBCLASS_WA,
	SUBCLASS_EQ,
	SUBCLASS_PE
} SUBCLASS_TYPE;

int GetControler(void)
{
	return scwa.songno;
}

void SetControler(int num)
{
	scwa.songno = num;
}

void SetControlerMax(int num)
{
	scwa.songmax = num;
}

static BOOL SubclassWinampCommon(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam, SUBCLASS_TYPE type)
{
	WPARAM lwParam;
	switch (uMsg)
	{
		case WM_WA_IPC:
			switch (lParam)
			{
				case IPC_PLAYFILE:
				case IPC_STARTPLAY:
				case IPC_SETPLAYLISTPOS:	/* ???? */
					UnsubclassWinamp();
					break;
			}
			break;
		case WM_DROPFILES:
			if (type != SUBCLASS_PE)
			{
				UnsubclassWinamp();
			}
			break;
		case WM_LBUTTONDBLCLK:
			if (type == SUBCLASS_PE)
			{
				UnsubclassWinamp();
			}
			break;
		case WM_COPYDATA:
			if (((COPYDATASTRUCT *)lParam)->dwData == IPC_PLAYFILE)
				UnsubclassWinamp();
			break;
		case WM_CLOSE:
			if (type == SUBCLASS_WA)
			{
				UnsubclassWinamp();
				/* re-post */
				PostMessage(hwnd, uMsg, wParam, lParam);
				return FALSE;
			}
			break;
		case WM_WA_MPEG_EOF:
			if (type == SUBCLASS_WA)
			{
				scwa.songno += 1;
				if (scwa.songno > scwa.songmax)
				{
					scwa.songno = scwa.songmax;
					UnsubclassWinamp();
					PostMessage(hwnd, WM_COMMAND, WINAMP_BUTTON5, 0);
				}
				else
				{
					PostMessage(hwnd, WM_COMMAND, WINAMP_BUTTON2, 0);
				}
				return FALSE;
			}
			break;
		case WM_COMMAND:
		case WM_SYSCOMMAND:
			lwParam = LOWORD(wParam);
			if (0x8d70 <= lwParam && lwParam < 40000)
			{
				/* BOOKMARK */
				UnsubclassWinamp();
				break;
			}
			switch (lwParam)
			{
				case WINAMP_FILE_QUIT:
					UnsubclassWinamp();
					/* re-post */
					PostMessage(hwnd, uMsg, wParam, lParam);
					return FALSE;
				case WINAMP_JUMP:			/* ???? */
				case WINAMP_JUMPFILE:		/* ???? */
				case IDC_PLAYLIST_PLAY:
				case WINAMP_FILE_PLAY:		/* open file */
				case WINAMP_FILE_LOC:		/* open URL */
				case WINAMP_FILE_DIR:		/* open folder */
				case WINAMP_BUTTON2:		/* play current file */
				case WINAMP_BUTTON4_CTRL:	/* stop list */
				case WINAMP_BUTTON1_CTRL:	/* play top of list */
				case WINAMP_BUTTON5_CTRL:	/* play end of list */
					UnsubclassWinamp();
					break;
				case WINAMP_JUMP10BACK:
				case WINAMP_BUTTON1:	/* back button */
					if (scwa.songno <= 1)
					{
						scwa.songno = 1;
						UnsubclassWinamp();
						PostMessage(hwnd, WM_COMMAND, WINAMP_BUTTON1, 0);
					}
					scwa.songno -= (lwParam == WINAMP_JUMP10BACK) ? 10 : 1;
					if (scwa.songno < 1) scwa.songno = 1;
					if (SendMessage(scwa.hwndWA, WM_WA_IPC,0,IPC_ISPLAYING))
						PostMessage(hwnd, WM_COMMAND, WINAMP_BUTTON2, 0);
					return FALSE;
				case WINAMP_BUTTON5:	/* fwd button */
				case WINAMP_JUMP10FWD:
					if (scwa.songno >= scwa.songmax)
					{
						scwa.songno = scwa.songmax;
						UnsubclassWinamp();
						PostMessage(hwnd, WM_COMMAND, WINAMP_BUTTON5, 0);
						return FALSE;
					}
					scwa.songno += (lwParam == WINAMP_JUMP10FWD) ? 10 : 1;
					if (scwa.songno > scwa.songmax) scwa.songno = scwa.songmax;
					if (SendMessage(scwa.hwndWA, WM_WA_IPC,0,IPC_ISPLAYING))
						PostMessage(hwnd, WM_COMMAND, WINAMP_BUTTON2, 0);
					return FALSE;
				}
			break;
	}
	return TRUE;
}

static LRESULT CALLBACK SubclassWAProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	WNDPROC wpOrig = scwa.wpOrigWA;
	if (!scwa.enable || SubclassWinampCommon(hwnd, uMsg, wParam, lParam, SUBCLASS_WA))
		return CallWindowProc(wpOrig, hwnd, uMsg, wParam, lParam); 
	else
		return 0;
}

static LRESULT CALLBACK SubclassEQProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	WNDPROC wpOrig = scwa.wpOrigEQ;
	if (!scwa.enable || SubclassWinampCommon(hwnd, uMsg, wParam, lParam, SUBCLASS_EQ))
		return CallWindowProc(wpOrig, hwnd, uMsg, wParam, lParam); 
	else
		return 0;
}

static LRESULT CALLBACK SubclassPEProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	WNDPROC wpOrig = scwa.wpOrigPE;
	if (!scwa.enable || SubclassWinampCommon(hwnd, uMsg, wParam, lParam, SUBCLASS_PE))
		return CallWindowProc(wpOrig, hwnd, uMsg, wParam, lParam); 
	else
		return 0;
}

void UnsubclassWinamp(void)
{
	if (CRITICAL_SECTION_ENTER(scwa_lock))
	{
		if (scwa.enable)
		{
			/* OutputDebugString("Unsubclasssed amp.\n"); */
			scwa.enable = FALSE;
			if (scwa.hwndWA != NULL)
			{
				SetWindowLong(scwa.hwndWA, GWL_WNDPROC, (LONG)scwa.wpOrigWA);
				scwa.hwndWA = NULL;
			}
			if (scwa.hwndEQ != NULL)
			{
				SetWindowLong(scwa.hwndEQ, GWL_WNDPROC, (LONG)scwa.wpOrigEQ);
				scwa.hwndEQ = NULL;
			}
			if (scwa.hwndPE != NULL)
			{
				SetWindowLong(scwa.hwndPE, GWL_WNDPROC, (LONG)scwa.wpOrigPE);
				scwa.hwndPE = NULL;
			}
		}
		CRITICAL_SECTION_LEAVE(scwa_lock);
	}
}

typedef struct
{
	DWORD dwThreadId;
	LPCSTR lpszClassName;
	HWND hwndRet;
} FINDTHREADWINDOW_WORK;

static BOOL CALLBACK FindThreadWindowsProc(HWND hwnd, LPARAM lParam)
{
	FINDTHREADWINDOW_WORK *pWork = (FINDTHREADWINDOW_WORK *)lParam;
	if (GetWindowThreadProcessId(hwnd, NULL) == pWork->dwThreadId)
	{
#define MAX_CLASS_NAME MAX_PATH
		CHAR szClassName[MAX_CLASS_NAME];
		if (GetClassName(hwnd, szClassName, MAX_CLASS_NAME))
		{
			if (lstrcmp(szClassName, pWork->lpszClassName) == 0)
			{
				pWork->hwndRet = hwnd;
				return FALSE;
			}
		}
	}
	return TRUE;
}

static HWND FindThreadWindow(LPCSTR lpszClassName, DWORD dwThreadId)
{
	FINDTHREADWINDOW_WORK fwww;
	fwww.dwThreadId = dwThreadId;
	fwww.lpszClassName = lpszClassName;
	fwww.hwndRet = NULL;
	EnumWindows(FindThreadWindowsProc, (LONG)&fwww);
	return fwww.hwndRet;
}

void SubclassWinamp(void)
{
	if (CRITICAL_SECTION_ENTER(scwa_lock))
	{
		HWND hwndWA = winampGetInModule2()->hMainWindow;
		if (!scwa.enable && hwndWA != NULL)
		{
			DWORD dwThreadId;
			scwa.hwndWA = hwndWA;
			dwThreadId = GetWindowThreadProcessId(scwa.hwndWA, NULL);
			scwa.wpOrigWA = (WNDPROC)SetWindowLong(scwa.hwndWA, GWL_WNDPROC, (LONG)SubclassWAProc);
			scwa.hwndEQ = FindThreadWindow("Winamp EQ", dwThreadId);
			if (scwa.hwndEQ != NULL)
			{
				scwa.wpOrigEQ = (WNDPROC)SetWindowLong(scwa.hwndEQ, GWL_WNDPROC, (LONG)SubclassEQProc);
			}
			scwa.hwndPE = FindThreadWindow("Winamp PE", dwThreadId);
			if (scwa.hwndPE != NULL)
			{
				scwa.wpOrigPE = (WNDPROC)SetWindowLong(scwa.hwndPE, GWL_WNDPROC, (LONG)SubclassPEProc);
			}
			/* OutputDebugString("Subclasssed amp.\n"); */
			scwa.enable = TRUE;
		}
		CRITICAL_SECTION_LEAVE(scwa_lock);
	}
}
